@@ -255,7 +255,7 @@ retry:
 		  break;
 		}
 	    }
-	  /*FALLTHRU*/
+	  fallthrough;
 	case STATUS_INVALID_DEVICE_REQUEST:
 	case STATUS_INVALID_PARAMETER:
 	case STATUS_INVALID_HANDLE:
@@ -265,6 +265,7 @@ retry:
 	      len = (size_t) -1;
 	      break;
 	    }
+	  fallthrough;
 	default:
 	  __seterrno_from_nt_status (status);
 	  len = (size_t) -1;
@@ -949,9 +950,9 @@ fhandler_base::write (const void *ptr, size_t len)
 	    }
 
 	  /* We've got a buffer-full, or we're out of data.  Write it out */
-	  int nbytes;
-	  int want = buf_ptr - buf;
-	  if ((nbytes = raw_write (buf, want)) == want)
+	  ssize_t nbytes;
+	  ptrdiff_t want = buf_ptr - buf;
+	  if ((nbytes = raw_write (buf, (size_t) want)) == want)
 	    {
 	      /* Keep track of how much written not counting additional \r's */
 	      res = data - (char *)ptr;
