@@ -491,7 +491,7 @@ fhandler_base::fstat_helper (struct stat *buf)
       else
 	{
 	  buf->st_dev = buf->st_rdev = dev ();
-	  buf->st_mode |= dev ().mode () & S_IFMT;
+	  buf->st_mode = dev ().mode ();
 	  buf->st_size = 0;
 	}
     }
@@ -696,12 +696,12 @@ fhandler_base::fstatvfs_by_handle (HANDLE fh, struct statvfs *sfs)
 	}
       else
 	debug_printf ("%y = NtQueryVolumeInformationFile"
-		      "(%S, FileFsSizeInformation)", 
+		      "(%S, FileFsSizeInformation)",
 		      status, pc.get_nt_native_path ());
     }
   else
     debug_printf ("%y = NtQueryVolumeInformationFile"
-		  "(%S, FileFsFullSizeInformation)", 
+		  "(%S, FileFsFullSizeInformation)",
 		  status, pc.get_nt_native_path ());
   return ret;
 }
@@ -827,7 +827,7 @@ fhandler_disk_file::fchmod (mode_t mode)
       HANDLE fh;
 
       if (NT_SUCCESS (NtOpenFile (&fh, FILE_WRITE_ATTRIBUTES,
-      				  pc.init_reopen_attr (attr, get_handle ()),
+				  pc.init_reopen_attr (attr, get_handle ()),
 				  &io, FILE_SHARE_VALID_FLAGS,
 				  FILE_OPEN_FOR_BACKUP_INTENT)))
 	{
